// CopperTop (C++ Console) : main project file.

#include "bass.h"
#include "bassasio.h"
#include <math.h>
#include <stdio.h>

/*****************************************************************************/

//#define TEST_TONES // use this option to generate a separate test tone at each output
#define SAMPLE_RATE_HZ 48000 // The Audigy always runs at 48 kHz, but re-samples any incoming data
// so it will read a 44k1 stream transparently. You can use 44k1 directly
// with the X-Fi.
#define INPUT_CHAN_TYPE 1
#define OUTPUT_CHAN_TYPE 0
#define NO_FLAGS 0
#define NUM_OP_CHANS 6 // as many outputs you need e.g. three way stereo crossover
#define ASIO_START_CHAN 2 // for the Audigy, by default ASIO channels 2-8 are mapped to physical outputs
// but this can be changed in the Kx router
#define INPUT_VOL 1.0
#define OUTPUT_VOL 1.0
#define STEREO 2
#define BUF_LENGTH 65536 // you can be generous with a PC. N.B This doesn't define the latency.
#define PI 3.141592654

/*****************************************************************************/

// create buffers for incoming stereo samples
short pbuf_l [BUF_LENGTH];
short pbuf_r [BUF_LENGTH];
// pointers for circular buffer
int wr_ptr = 0;
int rd_ptr = 0;

/*****************************************************************************/

// RECORD - buffer the recorded data
BOOL CALLBACK RecordingCallback(HRECORD handle, short *r_buffer, DWORD length, DWORD user)
{
static unsigned int ii;
//printf ("R %d\n", length); // just for debugging
/* always transfer from low level sampled array into our own circular buffer */
for (ii=0; ii < length/2; ii += 2) // 'length' is in bytes
    {
    pbuf_l [wr_ptr] = r_buffer [ii];
    pbuf_r [wr_ptr] = r_buffer [ii + 1];
    if (wr_ptr < BUF_LENGTH-1)
        {
        wr_ptr ++;
        }
    else
        {
        wr_ptr = 0;
        }
    }
    return TRUE;
}

/*****************************************************************************/

// ASIO playback function
DWORD CALLBACK AsioOutProc(BOOL input, DWORD channel, float *buffer, DWORD length, void *user)
{
unsigned int ii, ch;
static float phase [NUM_OP_CHANS] = {0,0,0,0,0,0}; // if using test tones
//printf ("P %d\n", length); // just for debugging

#ifdef TEST_TONES

for (ii=0; ii < length / 4; ii += NUM_OP_CHANS)
    {
    for (ch=0; ch < NUM_OP_CHANS; ch ++)
        {
        buffer [ii + ch] = (float) sin ((float) phase [ch]) /10;
        phase [ch] += (float)0.01 * (ch + 1);
        if (phase [ch] > 2*PI)
        {
        phase [ch] -= 2*PI;
        }
    }
}

#else // as a demonstration, copy incoming audio to outputs

for (ii=0; ii < length / 4; ii += NUM_OP_CHANS)
    {
    for (ch = 0; ch < NUM_OP_CHANS; ch ++)
        {
        // feed stereo input left to even channels, right to odd
        if (ch % 2 == 0)
            {
            buffer [ii + ch] = (float) pbuf_l [rd_ptr] / 32768;
            }
        else
            {
            buffer [ii + ch] = (float) pbuf_r [rd_ptr] / 32768;
            }
        }
    if (rd_ptr < BUF_LENGTH-1)
        {
        rd_ptr ++;
        }
    else
        {
        rd_ptr = 0;
        }
    }

#endif

return length;
}

/*****************************************************************************/

int main(void)
{
int a;
BASS_DEVICEINFO device_info;
BASS_ASIO_DEVICEINFO asio_device_info;
char *name;
//variables for recording and playback device assignment
int rec_dev, pb_dev, rec_ip;

printf ("List of ASIO output devices:\n");
for (a = 0;BASS_ASIO_GetDeviceInfo(a, &asio_device_info); a++)
    {
    printf("dev %d: %s\ndriver: %s\n",a, asio_device_info.name, asio_device_info.driver);
    if (BASS_ASIO_Init(a,0))
        {
        BASS_ASIO_CHANNELINFO i;
        int b;
        for (b=0;BASS_ASIO_ChannelGetInfo(1,b,&i);b++)
            {
            printf ("\tin %d: %s (group %d, format %d)\n",b,i.name,i.group,i.format);
            }
        for (b=0;BASS_ASIO_ChannelGetInfo(0,b,&i);b++)
        {
        printf("\tout %d: %s (group %d, format %d)\n",b,i.name,i.group,i.format);
        }
    }
    BASS_ASIO_Free();
    }

printf ("Use which ASIO playback device ? ");
scanf ("%d", &pb_dev);
printf ("\n");

// list recording devices
printf ("List of recording devices:\n");
for (a=0; BASS_RecordGetDeviceInfo(a, &device_info); a++)
    {
    printf ("%d: Device:%s Driver:%s\n", a, device_info.name, device_info.driver);
    }

printf ("Use which recording device ? ");
scanf ("%d", &rec_dev);
printf ("\n");

BASS_RecordInit(rec_dev);

printf ("List of recording inputs:\n");
for (a = 0; name = (char*)BASS_RecordGetInputName(a); a++)
    {
    float vol;
    int s = BASS_RecordGetInput(a, &vol);
    printf("%d: %s \n", a, name);
    }

printf ("Use which recording input ? ");
scanf ("%d", &rec_ip);
printf ("\n");

BASS_RecordSetInput (rec_ip, NO_FLAGS, INPUT_VOL);
// initialise the desired ASIO output device
BASS_ASIO_Init(pb_dev, NO_FLAGS);
// set the sample rate
BASS_ASIO_SetRate ((double)SAMPLE_RATE_HZ);
// enable first ASIO output channel
//BASS_ASIO_ChannelEnable(OUTPUT_CHAN_TYPE, ASIO_START_CHAN, &AsioOutProc,0);
BASS_ASIO_ChannelEnable(OUTPUT_CHAN_TYPE, ASIO_START_CHAN, (ASIOPROC*) AsioOutProc,0);
// using floating-point
BASS_ASIO_ChannelSetFormat(OUTPUT_CHAN_TYPE, ASIO_START_CHAN, BASS_ASIO_FORMAT_FLOAT);
// set the output volume for this channel
BASS_ASIO_ChannelSetVolume (OUTPUT_CHAN_TYPE, ASIO_START_CHAN, (double)OUTPUT_VOL);

// and join the next N ASIO output channels to it to make one 6 channel group
for (a=1; a < NUM_OP_CHANS; a++)
    {
    BASS_ASIO_ChannelJoin(OUTPUT_CHAN_TYPE, ASIO_START_CHAN + a, ASIO_START_CHAN);
    }

//* kick off an audio recording */
// start recording @ SAMPLE_RATEhz 16-bit stereo
BASS_RecordStart(SAMPLE_RATE_HZ, STEREO, NO_FLAGS,(RECORDPROC*) &RecordingCallback,0);

// Put a pause in here to ensure that enough recorded samples have
// built up before starting playback. You may wish to think carefully about this
// to reduce latency
while (wr_ptr - rd_ptr < 4096);

// start playing back over ASIO
BASS_ASIO_Start (4096); // You can force a specific buffer length with a non-zero argument
// loop forever, and let the callbacks do all the work
while (1)
    {
    Sleep (500);
    }
    return 1;
    }
